#include "map.h"
#include "fov.h"
#include "display.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#define FOVRADIUS	100

/* Global variables ----------------------------------------------- */

MAP map(3);
int pX=25;
int pY=20;
unsigned radius=20;

/* Callbacks ------------------------------------------------------ */

/**
 * Function called by libfov to apply light to a cell.
 *
 * \param map Pointer to map data structure passed to function such as
 *            fov_circle.
 * \param x   Absolute x-axis position of cell.
 * \param y   Absolute x-axis position of cell.
 * \param dx  Offset of cell from source cell on x-axis.
 * \param dy  Offset of cell from source cell on y-axis.
 * \param src Pointer to source data structure passed to function such
 *            as fov_circle.
 */
void apply(void *map, int x, int y, int dx, int dy, void *src) {
	if (((MAP *)map)->onMap(x, y))
		((MAP *)map)->setSeen(x, y);
}

/**
 * Function called by libfov to determine whether light can pass
 * through a cell. Return zero if light can pass though the cell at
 * (x,y), non-zero if it cannot.
 *
 * \param map Pointer to map data structure passed to function such as
 *            fov_circle.
 * \param x   Absolute x-axis position of cell.
 * \param y   Absolute x-axis position of cell.
 */
int opaque(void *map, int x, int y) {
	return ((MAP *)map)->blockLOS(x, y);
}

/**
 * Offset player coordinates by (dx,dy). e.g. player_move(1,0) moves
 * the player right one cell.
 */
void player_move(int dx, int dy) {
	unsigned newx = pX + dx;
	unsigned newy = pY + dy;
	if (map.onMap(newx, newy)) {
		pX = newx;
		pY = newy;
	}
}

/**
 * Redraw the screen. Called in a loop to create the output.
 */
void redraw() {
	display_clear();

	printf("------------------------------------------------\n");

	static struct timeval tp;
	static double t1, t2, e;
	gettimeofday(&tp, NULL);
	t1=(double)tp.tv_sec+(1.0e-6)*tp.tv_usec;

	/* Cause libfov to mark lit cells using the two callbacks given:
	 *  - opaque, which is used to determine information about your
	 *    map (a query you can define), and
	 *  - apply, which is used to modify your map once libfov
	 *    determines that a particular cell is lit (a command you can
	 *    define).
	 *
	 * In this call, the light source is at (pX,pY). We pass NULL for
	 * the source pointer parameter because we have no data structure
	 * defining our light source in this example. You may want to pass
	 * a pointer to a data structure holding your light source if, for
	 * example, you wanted different coloured lights. In that case the
	 * colour of the cell would depend upon the colour of the source
	 * and you would have to extract this colour from your data
	 * structure in each call to apply.
	 */
	fov_circle(&map, NULL, pX, pY, radius, opaque, apply);

	gettimeofday(&tp, NULL);
	t2=(double)tp.tv_sec+(1.0e-6)*tp.tv_usec;

	e = t2-t1;

	printf("Time: %e\n", e);

	map.display();
	display_put_char('@', pX, pY, 0x00, 0xFF, 0x00);

	display_refresh();
}

/**
 * Handle a keypress. Callback used by display_event_loop.
 */
void keypressed(int key) {
	switch (key) {
	case KEY_UP:    player_move( 0, -1); break;
	case KEY_DOWN:  player_move( 0,  1); break;
	case KEY_LEFT:  player_move(-1,  0); break;
	case KEY_RIGHT: player_move( 1,  0); break;
	case SDLK_EQUALS: radius++; break;
	case SDLK_MINUS:  radius--; if (radius <= 0) radius = 1; break;
	case SDLK_p: 
		printf("Precalculated circular limit\n");
		fov_set_limit_circular_precalculated();
		break;
	case SDLK_c: 
		printf("Circular limit\n");
		fov_set_limit_circular();
		break;
	case SDLK_s:
		printf("Square limit\n");
		fov_set_limit_square();
		break;
	case SDLK_o:
		printf("Octagonal limit\n");
		fov_set_limit_octagonal();
		break;
	case SDLK_a:
		printf("Applying to opaque tiles\n");
		fov_set_opaque_apply();
		break;
	case SDLK_n:
		printf("Not applying to opaque tiles\n");
		fov_set_opaque_noapply();
		break;
	case KEY_q:
	case KEY_ESCAPE:
		display_exit();
		fov_free();
		exit(0);
		break;
	default:
		break;
	}
	redraw();
}

/**
 * Program entry point.
 */
int main(void) {
	display_init();
	redraw();
	display_event_loop(keypressed);
	display_exit();
	fov_free();
	return 0;
}
